// AcquireContinuous.cpp
//


// include files to use the PYLON API
#include <pylon/PylonIncludes.h>
#include <stdlib.h>

using namespace Pylon;

#if defined( USE_1394 )
// settings to use  Basler 1394 cameras
#include <pylon/1394/Basler1394Camera.h>
typedef Pylon::CBasler1394Camera Camera_t;
using namespace Basler_IIDC1394CameraParams;
using namespace Basler_IIDC1394StreamParams;
#elif defined ( USE_GIGE )
// settings to use Basler GigE cameras
#include <pylon/gige/BaslerGigECamera.h>
typedef Pylon::CBaslerGigECamera Camera_t;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;
#else
#error Camera type is not specified. For example, define USE_GIGE for using GigE cameras
#endif

// namespace for using cout
using namespace std;




class CGrabBuffer
{
    public:
        CGrabBuffer(const size_t ImageSize);
        ~CGrabBuffer();
        uint8_t* GetBufferPointer(void) { return m_pBuffer; }
        StreamBufferHandle GetBufferHandle(void) { return m_hBuffer; }
        void SetBufferHandle(StreamBufferHandle hBuffer) { m_hBuffer = hBuffer; };

    protected:
        uint8_t *m_pBuffer;
        StreamBufferHandle m_hBuffer;
};

// Constructor allocates the image buffer
CGrabBuffer::CGrabBuffer(const size_t ImageSize):
        m_pBuffer(NULL)
{
    m_pBuffer = new uint8_t[ ImageSize ];
    if (NULL == m_pBuffer)
    {
        GenICam::GenericException e("Not enough memory to allocate image buffer", __FILE__, __LINE__);
        throw e;
    }
}

// Freeing the memory
CGrabBuffer::~CGrabBuffer()
{
    if (NULL != m_pBuffer)
        delete[] m_pBuffer;
}


// Buffers for grabbing
static const uint32_t c_nBuffers = 10;
// Number of images to be grabbed
static const uint32_t c_ImagesToGrab = 100;

int main(int argc, char* argv[])
{
    // Automagically call PylonInitialize and PylonTerminate to ensure the pylon runtime system
    // is initialized during lifetime of this object.
    Pylon::PylonAutoInitTerm autoInitTerm;

    try
    {
        // Get the transport layer factory
        CTlFactory& TlFactory = CTlFactory::GetInstance();

        // Create the transport layer object needed to enumerate or
        // create a camera object of type Camera_t::DeviceClass()
        ITransportLayer *pTl = TlFactory.CreateTl(Camera_t::DeviceClass());

        // Exit application if the specific transport layer is not available
        if (! pTl)
        {
            cerr << "Failed to create transport layer!" << endl;
            return 1;
        }

        // Get all attached cameras and exit application if no camera is found
        DeviceInfoList_t devices;
        if (0 == pTl->EnumerateDevices(devices))
        {
            cerr << "No camera present!" << endl;
            return 1;
        }

        // Create the camera object of the first available camera
        // The camera object is used to set and get all available
        // camera features.
        Camera_t Camera(pTl->CreateDevice(devices[ 0 ]));

        // Get the first stream grabber object of the selected camera
        Camera_t::StreamGrabber_t StreamGrabber(Camera.GetStreamGrabber(0));


        // Open the camera
        Camera.Open();

        // Open the stream grabber
        StreamGrabber.Open();


        // Set image format and AOI
        Camera.PixelFormat.SetValue(PixelFormat_Mono8);
        Camera.OffsetX.SetValue(0);
        Camera.OffsetY.SetValue(0);
        Camera.Width.SetValue(Camera.Width.GetMax());
        Camera.Height.SetValue(Camera.Height.GetMax());

        // Set the camera to continuous frame mode
        Camera.TriggerSelector.SetValue(TriggerSelector_AcquisitionStart);
        Camera.TriggerMode.SetValue(TriggerMode_Off);
        Camera.AcquisitionMode.SetValue(AcquisitionMode_Continuous);
        Camera.ExposureMode.SetValue(ExposureMode_Timed);
        Camera.ExposureTimeRaw.SetValue(100);

        // Get the image buffer size
        const size_t ImageSize = (size_t)(Camera.PayloadSize.GetValue());

        // We won't use image buffers greater than ImageSize
        StreamGrabber.MaxBufferSize.SetValue(ImageSize);

        // We won't queue more than c_nBuffers image buffer at a time
        StreamGrabber.MaxNumBuffer.SetValue(c_nBuffers);

        // Allocate all resources for grabbing. Critical parameters like image
        // size now must not be changed until FinishGrab() is called.
        StreamGrabber.PrepareGrab();

        // Buffers used for grabbing must be registered at the stream grabber.
        // The registration returns a handle to be used for queuing the buffer
        std::vector<CGrabBuffer*> BufferList;
        for (uint32_t i = 0; i < c_nBuffers; ++i)
        {
            CGrabBuffer *pGrabBuffer = new CGrabBuffer(ImageSize);
            pGrabBuffer->SetBufferHandle(StreamGrabber.RegisterBuffer(
                pGrabBuffer->GetBufferPointer(), ImageSize));

            // Put the grab buffer object into the buffer list
            BufferList.push_back(pGrabBuffer);
        }

        for (std::vector<CGrabBuffer*>::const_iterator x = BufferList.begin(); x != BufferList.end(); ++x)
        {
            // Put buffer into the grab queue for grabbing
            StreamGrabber.QueueBuffer((*x)->GetBufferHandle(), NULL);
        }

        // Let the camera acquire continuously images ( Acquisiton mode equals
        // Continuous! )
        Camera.AcquisitionStart.Execute();

        // Grab c_ImagesToGrab times
        for (int n = 0; n < c_ImagesToGrab; n++)
        {
            // Wait for the grabbed image with timeout of 3 seconds
            if (StreamGrabber.GetWaitObject().Wait(3000))
            {
                // Get the grab result from the grabber's result queue
                GrabResult Result;
                StreamGrabber.RetrieveResult(Result);

                if (Grabbed == Result.Status())
                {
                    // Grabbing was successful, process image
                    cout << "Image #" << n << " acquired!" << endl;
                    cout << "Size: " << Result.GetSizeX() << " x "
                    << Result.GetSizeY() << endl;

                    // Get the pointer to the image buffer
                    const uint8_t *pImageBuffer = (uint8_t *) Result.Buffer();
                    cout << "Gray value of first pixel: " << (uint32_t) pImageBuffer[0]
                    << endl << endl;

                    // Reuse the buffer for grabbing the next image
                    if (n < c_ImagesToGrab - c_nBuffers)
                        StreamGrabber.QueueBuffer(Result.Handle(), NULL);
                }
                else if (Failed == Result.Status())
                {
                    // Error Handling
                    cerr << "No image acquired!" << endl;
                    cerr << "Error code : 0x" << hex
                    << Result.GetErrorCode() << endl;
                    cerr << "Error description : "
                    << Result.GetErrorDescription() << endl;

                    // Reuse the buffer for grabbing the next image
                    if (n < c_ImagesToGrab - c_nBuffers)
                        StreamGrabber.QueueBuffer(Result.Handle(), NULL);
                }
            }
            else
            {
                // Timeout
                cerr << "Timeout occurred!" << endl;

                // Get the pending buffer back (It is not allowed to deregister
                // buffers when they are still queued)
                StreamGrabber.CancelGrab();

                // Get all buffers back
                for (GrabResult r; StreamGrabber.RetrieveResult(r););

                // Cancel loop
                break;
            }
        }

        // Stop acquisition
        Camera.AcquisitionStop.Execute();

        // clean up

        // It is necessary to deregister buffer before freeing the memory
        for (std::vector<CGrabBuffer*>::iterator it = BufferList.begin(); it != BufferList.end(); it++)
        {
            StreamGrabber.DeregisterBuffer((*it)->GetBufferHandle());
            delete *it;
        }

        // Free all resources used for grabbing
        StreamGrabber.FinishGrab();

        // Close stream grabber
        StreamGrabber.Close();

        // Close camera
        Camera.Close();

    }
    catch (GenICam::GenericException &e)
    {
        // Error handling
        cerr << "An exception occurred!" << endl

        << e.GetDescription() << endl;
        return 1;
    }
    // Quit application
    return 0;
}

