// AcquireSingleFrame_ChunkImage.cpp
//

// include files to use the PYLON API
#include <pylon/PylonIncludes.h>
using namespace Pylon;

#if defined( USE_1394 )
// settings to use  Basler 1394 cameras
#include <pylon/1394/Basler1394Camera.h>
typedef Pylon::CBasler1394Camera Camera_t;
using namespace Basler_IIDC1394CameraParams;
using namespace Basler_IIDC1394StreamParams;
#elif defined ( USE_GIGE )
// settings to use Basler GigE cameras
#include <pylon/gige/BaslerGigECamera.h>
typedef Pylon::CBaslerGigECamera Camera_t;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;
#else
#error Camera type is not specified. For example, define USE_GIGE for using GigE cameras
#endif


// namespace for using cout
using namespace std;


int main(int argc, char* argv[])
{
    // Automagically call PylonInitialize and PylonTerminate to ensure the pylon runtime system
    // is initialized during lifetime of this object.
    Pylon::PylonAutoInitTerm autoInitTerm;

    try
    {
        // Get the transport layer factory
        CTlFactory& TlFactory = CTlFactory::GetInstance();

        // Create the transport layer object needed to enumerate or
        // create a camera object of type Camera_t::DeviceClass()
        ITransportLayer *pTl = TlFactory.CreateTl(Camera_t::DeviceClass());

        // Exit application if the specific transport layer is not available
        if (! pTl)
        {
            cerr << "Failed to create transport layer!" << endl;
            return 1;
        }

        // Get all attached cameras and exit application if no camera is found
        DeviceInfoList_t devices;
        if (0 == pTl->EnumerateDevices(devices))
        {
            cerr << "No camera present!" << endl;
            return 1;
        }

        // Create the camera object of the first available camera
        // The camera object is used to set and get all available
        // camera features.
        Camera_t Camera(pTl->CreateDevice(devices[ 0 ]));

        // Get the first stream grabber object of the selected camera
        Camera_t::StreamGrabber_t StreamGrabber(Camera.GetStreamGrabber(0));

        // Open the camera
        Camera.Open();

        // Open the stream grabber
        StreamGrabber.Open();

        // Set image format and AOI
        Camera.PixelFormat.SetValue(PixelFormat_Mono8);
        Camera.OffsetX.SetValue(0);
        Camera.OffsetY.SetValue(0);
        Camera.Width.SetValue(Camera.Width.GetMax());
        Camera.Height.SetValue(Camera.Height.GetMax());

        // Set the camera to single frame mode
        Camera.TriggerSelector.SetValue(TriggerSelector_AcquisitionStart);
        Camera.TriggerMode.SetValue(TriggerMode_Off);
        Camera.AcquisitionMode.SetValue(AcquisitionMode_SingleFrame);
        Camera.ExposureMode.SetValue(ExposureMode_Timed);
        Camera.ExposureTimeRaw.SetValue(100);

        // Enable chunks in general
        if (GenApi::IsWritable(Camera.ChunkModeActive))
        {
            Camera.ChunkModeActive.SetValue(true);
        }
        else
        {
            cerr << "The camera doesn't support chunk features" << endl;
            return 1;
        }

        // Enable time stamp chunks
        Camera.ChunkSelector.SetValue(ChunkSelector_Timestamp);
        Camera.ChunkEnable.SetValue(true);

        // Enable frame counter chunks
        Camera.ChunkSelector.SetValue(ChunkSelector_Framecounter);
        Camera.ChunkEnable.SetValue(true);

        // Enable crc chunks
        Camera.ChunkSelector.SetValue(ChunkSelector_PayloadCRC16);
        Camera.ChunkEnable.SetValue(true);

        // Create an image buffer
        const size_t ImageSize = (size_t)(Camera.PayloadSize.GetValue());
        uint8_t *const pBuffer = new uint8_t[ ImageSize ];

        // We won't use image buffers greater than ImageSize
        StreamGrabber.MaxBufferSize.SetValue(ImageSize);

        // We won't queue more than one image buffer at a time
        StreamGrabber.MaxNumBuffer.SetValue(1);

        // allocate resources related to grabbing
        StreamGrabber.PrepareGrab();

        // subscribe the buffer
        const StreamBufferHandle hBuffer =
            StreamGrabber.RegisterBuffer(pBuffer, ImageSize);

        // Put buffer into the grab queue for being filled with data
        StreamGrabber.QueueBuffer(hBuffer, NULL);

        // Let the camera acquire one image
        Camera.AcquisitionMode.SetValue(AcquisitionMode_SingleFrame);
        Camera.AcquisitionStart.Execute();


        GrabResult Result;
        // Wait for the buffer to be filled
        if (StreamGrabber.GetWaitObject().Wait(3000))
        {
            // Get the grab result from the grabber's result queue
            StreamGrabber.RetrieveResult(Result);

        }
        else
        {
            // Timeout
            cerr << "Timeout occurred!" << endl;
            return 1;
        }

        if (! Result.Succeeded())
        {
            // Error Handling
            cerr << "No image acquired!" << endl;
            cerr << "Error code : 0x" << hex
            << Result.GetErrorCode() << endl;
            cerr << "Error description : "
            << Result.GetErrorDescription() << endl;
            return 1;
        }

        // Check if a buffer containing chunk data has been received
        if (PayloadType_ChunkData != Result.GetPayloadType())
        {
            cerr << "Unexpected payload type received" << endl;
            return 1;
        }

        // Create ChunkParser
        IChunkParser* pChunkParser = Camera.CreateChunkParser();
        if (NULL == pChunkParser)
        {
            cerr << "There is no chunk parser available for this device type!" << endl;
            return 1;
        }


        // Attach image buffer with chunk data to the parser. The parser extracts
        // the included data from the chunk.
        pChunkParser->AttachBuffer(Result.Buffer(), Result.GetPayloadSize());

        // Since we have activated the CRC feature, we can check
        // the integrity of the buffer first.
        // Note: Enabling the CRC feature is not a prerequisite for using
        // chunks. Chunks can also be handled when the CRC feature is deactivated.
        if (pChunkParser->HasCRC() &&
                pChunkParser->CheckCRC() == false)
        {
            cerr << "Image was damaged!" << endl;
            return 1;
        }

        // Access the chunk data.
        // Before accessing the chunk data, it should be checked
        // if the chunk is readable. When it is readable, the buffer
        // contains the requested chunk data.
        if (IsReadable(Camera.ChunkTimestamp))
            cout << "TimeStamp : " << Camera.ChunkTimestamp.GetValue() << endl;

        if (IsReadable(Camera.ChunkFramecounter))
            cout << "FrameCounter : " << Camera.ChunkFramecounter.GetValue() << endl;

        // Detach buffer from adapter.
        // The buffer should always be detached before reusing the buffer for grabbing!
        // After detaching the buffer, the chunk data is no longer accessible.
        pChunkParser->DetachBuffer();

        // clean up

        // Destroy the chunk parser
        Camera.DestroyChunkParser(pChunkParser);

        // It is necessary to deregister buffer before freeing the memory
        StreamGrabber.DeregisterBuffer(hBuffer);

        // Free all resources used for grabbing
        StreamGrabber.FinishGrab();

        // Disable chunk mode
        Camera.ChunkModeActive.SetValue(false);

        // Close stream grabber
        StreamGrabber.Close();

        // Close camera
        Camera.Close();

        // Free memory of image buffer
        delete[] pBuffer;
    }
    catch (GenICam::GenericException &e)
    {
        // Error handling
        cerr << "An exception occurred!" << endl
        << e.GetDescription() << endl;
        return 1;
    }

    // Quit application
    return 0;
}
