// AcquireSingleFrame_Gen.cpp
//


#include <pylon/PylonIncludes.h>

// namespace for using cout
using namespace std;

// namespace for using PYLON
using namespace Pylon;
using namespace GenApi;
using namespace GenICam;


int main(int argc, char* argv[])
{
    // Automagically call PylonInitialize and PylonTerminate to ensure the pylon runtime system
    // is initialized during lifetime of this object.
    Pylon::PylonAutoInitTerm autoInitTerm;

    try
    {
        // Get the transport layer factory
        CTlFactory& TlFactory = CTlFactory::GetInstance();

        // Get all attached cameras and exit application if no camera is found
        DeviceInfoList_t devices;
        if (0 == TlFactory.EnumerateDevices(devices))
        {
            cerr << "No camera present!" << endl;
            return 1;
        }

        // Create the camera object of the first available camera
        // The camera object is used to set and get all available
        // camera features.
        IPylonDevice &Camera = *TlFactory.CreateDevice(devices[ 0 ]);

        // Get the first stream grabber object of the selected camera
        IStreamGrabber &StreamGrabber = *Camera.GetStreamGrabber(0);

        // Open the camera
        Camera.Open();

        // Open the grabber
        StreamGrabber.Open();

        // Get the stream parameter object
        INodeMap &StreamParameters  = *StreamGrabber.GetNodeMap();

        // Get the camera control object
        INodeMap &Control = *Camera.GetNodeMap();

        // Get AOI and payload objects size
        const CIntegerPtr Width = Control.GetNode("Width");
        const CIntegerPtr Height = Control.GetNode("Height");
        const CIntegerPtr PaylodSize = Control.GetNode("PayloadSize");

        // Set image format and AOI
        CEnumerationPtr(Control.GetNode("PixelFormat"))->FromString("Mono8");
        CIntegerPtr(Control.GetNode("OffsetX"))->SetValue(0);
        CIntegerPtr(Control.GetNode("OffsetY"))->SetValue(0);
        Width->SetValue(Width->GetMax());
        Height->SetValue(Height->GetMax());

        // Set the camera to single frame mode
        CEnumerationPtr(Control.GetNode("TriggerSelector"))->FromString("AcquisitionStart");
        CEnumerationPtr(Control.GetNode("TriggerMode"))->FromString("Off");
        CEnumerationPtr(Control.GetNode("AcquisitionMode"))->FromString("SingleFrame");
        CEnumerationPtr(Control.GetNode("ExposureMode"))->FromString("Timed");
        CIntegerPtr(Control.GetNode("ExposureTimeRaw"))->SetValue(100);

        // Create a image buffer
        const size_t ImageSize = (size_t)(PaylodSize->GetValue());
        uint8_t * const pBuffer = new uint8_t[ ImageSize ];

        // We won't use image buffers greater than ImageSize
        CIntegerPtr(StreamParameters.GetNode("MaxBufferSize"))->SetValue(ImageSize);

        // We won't queue more than one image buffer at a time
        CIntegerPtr(StreamParameters.GetNode("MaxNumBuffer"))->SetValue(1);

        // Allocate all resources for grabbing. Critical parameters like image
        // size now must not be changed until FinishGrab() is called.
        StreamGrabber.PrepareGrab();

        // Buffers used for grabbing must be registered at the stream grabber.
        // The registration returns a handle to be used for queuing the buffer
        const StreamBufferHandle hBuffer =
            StreamGrabber.RegisterBuffer(pBuffer, ImageSize);

        // Put buffer into the grab queue for grabbing
        StreamGrabber.QueueBuffer(hBuffer, NULL);

        // Grab 10 times
        const uint32_t numGrabs = 10;
        for (int n = 0; n < numGrabs; n++)
        {
            // Let the camera acquire one single image ( Acquisiton mode equals
            // SingleFrame! )
            CCommandPtr(Control.GetNode("AcquisitionStart"))->Execute();

            // Wait for grabbed image with timeout of 3 seconds
            if (StreamGrabber.GetWaitObject().Wait(3000))
            {
                // Get the grab result from the grabber's result queue
                GrabResult Result;
                StreamGrabber.RetrieveResult(Result);

                if (Result.Succeeded())
                {
                    // Grabbing was successful, process image
                    cout << "Image #" << n << " acquired!" << endl;
                    cout << "Size: " << Result.GetSizeX() << " x "
                    << Result.GetSizeY() << endl;

                    // Get the pointer to the image buffer
                    const uint8_t *pImageBuffer = (uint8_t *) Result.Buffer();
                    cout << "Gray value of first pixel: " << (uint32_t) pImageBuffer[0]
                    << endl << endl;

                    // Reuse the buffer for grabbing the next image
                    if (n < numGrabs - 1)
                        StreamGrabber.QueueBuffer(Result.Handle(), NULL);
                }
                else
                {
                    // Error Handling
                    cerr << "No image acquired!" << endl;
                    cerr << "Error code : 0x" << hex
                    << Result.GetErrorCode() << endl;
                    cerr << "Error description : "
                    << Result.GetErrorDescription() << endl;

                    // Cancel loop
                    break;
                }
            }
            else
            {
                // Timeout
                cerr << "Timeout occurred!" << endl;

                // Get the pending buffer back (It is not allowed to deregister
                // buffers when they are still queued)
                StreamGrabber.CancelGrab();

                // Get all buffers back
                for (GrabResult r; StreamGrabber.RetrieveResult(r););

                // Cancel loop
                break;
            }
        }
        // clean up

        // It is necessary to deregister buffer before freeing the memory
        StreamGrabber.DeregisterBuffer(hBuffer);

        // Free all resources used for grabbing
        StreamGrabber.FinishGrab();

        // Close stream grabber
        StreamGrabber.Close();

        // Close camera
        Camera.Close();

        // Delete camera object
        TlFactory.DestroyDevice(&Camera);

        // Free memory of image buffer
        delete[] pBuffer;
    }
    catch (GenICam::GenericException &e)
    {
        // Error handling
        cerr << "An exception occurred!" << endl
        << e.GetDescription() << endl;
        return 1;
    }

    // Quit application
    return 0;
}

