// CameraEvents.cpp
//
// This sample program illustrates how to receive and interpret event messages sent by the camera.
// The usage of the end-of-exposure event is shown.

// include files to use the PYLON API
#include <pylon/PylonIncludes.h>
using namespace Pylon;


#if defined( USE_1394 )
// settings to use  Basler 1394 cameras
#include <pylon/1394/Basler1394Camera.h>
typedef Pylon::CBasler1394Camera Camera_t;
using namespace Basler_IIDC1394CameraParams;
using namespace Basler_IIDC1394StreamParams;
#elif defined ( USE_GIGE )
// settings to use Basler GigE cameras
#include <pylon/gige/BaslerGigECamera.h>
typedef Pylon::CBaslerGigECamera Camera_t;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;
using namespace Basler_GigETLParams;
#else
#error Camera type is not specified. For example, define USE_GIGE for using GigE cameras
#endif


// namespace for using cout
using namespace std;


// Simple buffer class
class CGrabBuffer
{
    public:
        CGrabBuffer(const size_t ImageSize);
        ~CGrabBuffer();
        uint8_t* GetBufferPointer(void) { return pBuffer; }
        StreamBufferHandle GetBufferHandle(void) { return m_hBuffer; }
        void SetBufferHandle(StreamBufferHandle hBuffer) { m_hBuffer = hBuffer; };

    protected:
        uint8_t *pBuffer;
        StreamBufferHandle m_hBuffer;
};

// Constructor allocates the image buffer
CGrabBuffer::CGrabBuffer(const size_t ImageSize):
        pBuffer(NULL)
{
    pBuffer = new uint8_t[ ImageSize ];
    if (NULL == pBuffer)
    {
        throw GENERIC_EXCEPTION("Not enough memory to allocate image buffer");
    }
}

// Freeing the memory
CGrabBuffer::~CGrabBuffer()
{
    if (NULL != pBuffer)
        delete[] pBuffer;
}

// Number of buffers used for grabbing images and events
const uint32_t c_nImageBuffers = 10;
const uint32_t c_nEventBuffers = 20;

// Number of images to be grabbed
const uint32_t c_ImagesToGrab = 20;

// Member function of this class will be registered as callback
struct CallbackTarget
{
    CallbackTarget(Camera_t& Camera)
            : m_Camera(Camera)
    { }

    // will be fired when an end-of-exposure event occurres
    void EndOfExposureCallback(GenApi::INode* pNode)
    {
        try
        {
            cout << "The message contains an end-of-exposure event." << endl;
            cout << "Timestamp: " << m_Camera.ExposureEndEventTimestamp.GetValue() << " Frame number: " << m_Camera.ExposureEndEventFrameID.GetValue() << endl;
        }
        catch (GenICam::GenericException& e)
        {
            cerr << "Failed to get event information. Exception occurred:"
            << e.GetDescription() << endl;
        }
    }

    Camera_t& m_Camera;
};


int main(int argc, char* argv[])
{
    // Automagically call PylonInitialize and PylonTerminate to ensure the pylon runtime system
    // is initialized during lifetime of this object.
    Pylon::PylonAutoInitTerm autoInitTerm;

    try
    {
        //
        // Create a camera object
        //

        // Create the transport layer object needed to enumerate or
        // create a camera object of type Camera_t::DeviceClass()
        ITransportLayer *pTl = CTlFactory::GetInstance().CreateTl(Camera_t::DeviceClass());
        if (! pTl)
        {
            cerr << "Failed to create transport layer!" << endl;
            return 1;
        }

        // Discover all attached devices
        DeviceInfoList_t devices;
        if (0 == pTl->EnumerateDevices(devices))
        {
            cerr << "No camera present!" << endl;
            return 1;
        }

        // Create the camera object of the first available camera
        Camera_t Camera(pTl->CreateDevice(devices[ 0 ]));

        //
        // Open the camera
        //
        Camera.Open(Exclusive | Stream | Control | Event);

        //
        // Check if the device supports events
        //
        if (! GenApi::IsAvailable(Camera.EventSelector))
        {
            cerr << "Device doesn't support events" << endl;
            Camera.Close();
            return 1;
        }

        //
        // Get and prepare the stream grabber
        //

        // determine the image buffer size
        const int bufferSize = (int) Camera.PayloadSize.GetValue();

        // Get the stream grabber object for the first stream channel
        Camera_t::StreamGrabber_t StreamGrabber(Camera.GetStreamGrabber(0));

        // Open the grabber
        StreamGrabber.Open();


        //
        // Get and prepare the event grabber
        //

        // Create the event grabber
        Camera_t::EventGrabber_t EventGrabber(Camera.GetEventGrabber());

        // parametrize and open it
        EventGrabber.NumBuffer.SetValue(c_nEventBuffers);    // must be set before open is called!!
#if USE_GIGE
        // Enable resending of event messages when lossed messages are detected:
        // Loss of messages is detected by sending acknowledges for every event messages.
        // When the camera doesn't receive the acknowledge it will resend the message up to
        // 'RetryCount' times.
        EventGrabber.RetryCount = 3;
#endif
        EventGrabber.Open();

        //
        // Create the event adapter used for parsing the event messages
        //

        Pylon::IEventAdapter *pEventAdapter = Camera.CreateEventAdapter();
        if (! pEventAdapter)
        {
            cerr << "Failed to create an event adapter" << endl;
            return 1;
        }

        //
        // Setup the camera for grabbing images continuously
        //

        Camera.PixelFormat.SetValue(PixelFormat_Mono8);
        Camera.TriggerSelector.SetValue(TriggerSelector_AcquisitionStart);
        Camera.TriggerMode.SetValue(TriggerMode_Off);
        Camera.AcquisitionMode.SetValue(AcquisitionMode_Continuous);
        Camera.ExposureMode.SetValue(ExposureMode_Timed);
        Camera.ExposureTimeRaw.SetValue(100);

        // Setup the stream grabber
        StreamGrabber.MaxBufferSize.SetValue(bufferSize);
        StreamGrabber.MaxNumBuffer.SetValue(c_nImageBuffers);
        StreamGrabber.PrepareGrab();

        //
        // Enable the end of exposure event
        //

        Camera.EventSelector.SetValue(EventSelector_ExposureEnd);    // select the event
        Camera.EventNotification.SetValue(EventNotification_GenICamEvent);   // enable it

        //
        // Register a callback for the end of exposure event
        //
        CallbackTarget callbackTarget(Camera);

        // Register the callback for the ExposureEndEventTimestamp node.
        GenApi::CallbackHandleType hCb = GenApi::Register(
                                             Camera.ExposureEndEventTimestamp.GetNode(),
                                             callbackTarget,
                                             &CallbackTarget::EndOfExposureCallback);

        //
        // Create and register the image buffers
        //
        std::vector<CGrabBuffer*> BufferList;
        for (uint32_t i = 0; i < c_nImageBuffers; ++i)
        {
            CGrabBuffer *pGrabBuffer = new CGrabBuffer(bufferSize);
            pGrabBuffer->SetBufferHandle(StreamGrabber.RegisterBuffer(
                                             pGrabBuffer->GetBufferPointer(), bufferSize));

            // Put buffer into the grab queue for grabbing
            StreamGrabber.QueueBuffer(pGrabBuffer->GetBufferHandle(), NULL);

            // Put the grab buffer object into the buffer list
            BufferList.push_back(pGrabBuffer);
        }

        // Get the wait objects of the event grabber and the stream grabber.
        // Put the objects in a container to be able to wait for images and events
        // simulataneously.
        WaitObjects waitset;
        waitset.Add(EventGrabber.GetWaitObject());
        waitset.Add(StreamGrabber.GetWaitObject());

        // Let the camera acquire continuously images ( Acquisiton mode equals
        // Continuous! )
        Camera.AcquisitionStart.Execute();

        //
        // Loop until c_ImagesToGrab images are grabbed.
        // Incoming events and images are processed.
        //

        int count = 0;
        while (count < c_ImagesToGrab)
        {
            unsigned int i;

            // wait for an image or an event to occure (5 sec timeout)
            if (waitset.WaitForAny(5000, &i))
            {
                // got event or image
                switch (i)
                {
                    case 0: // event available, process it
                    {
                        // Retrieve the event result
                        EventResult EvResult;
                        if (EventGrabber.RetrieveEvent(EvResult))
                        {
                            if (EvResult.Succeeded())
                            {
                                cout << "Successfully got an event message!" << endl;
                                // To figure out the content of the event message, pass it to the event adapter.
                                // DeliverMessage will fire the registered callback when the buffer contains
                                // an end-of-exposure event
                                pEventAdapter->DeliverMessage(EvResult.Buffer, sizeof EvResult.Buffer);
                            }
                            else
                            {
                                cerr << "Error retrieving event:" << EvResult.ErrorDescription() << endl;
                            }
                        }
                    }
                    break;

                    case 1: // image available, process it
                    {
                        GrabResult GrResult;
                        if (StreamGrabber.RetrieveResult(GrResult))
                        {
                            if (GrResult.Succeeded())
                            {
                                count++;
                                // Grabbing was successful, process image
                                cout << "Grabbed image."  << endl;
                            }
                            else if (Failed == GrResult.Status())
                            {
                                // Error Handling
                                cerr << "Failed to grab image!" << endl;
                                cerr << "Error code : 0x" << hex
                                << GrResult.GetErrorCode() << endl;
                                cerr << "Error description : "
                                << GrResult.GetErrorDescription() << endl;
                            }

                            // Reuse the buffer for grabbing further images
                            if (count <= c_ImagesToGrab - c_nImageBuffers)
                                StreamGrabber.QueueBuffer(GrResult.Handle(), NULL);
                        }
                    }
                    break;
                }  // switch
            }  // if WaitForAny
            else
            {
                // Timeout
                cerr << "Timeout occurred!" << endl;

                //
                // Get the pending image buffers back (It is not allowed to deregister
                // buffers when they are still queued)
                //

                // Release pending buffers
                StreamGrabber.CancelGrab();

                // Get all buffers back
                for (GrabResult r; StreamGrabber.RetrieveResult(r););

                // Cancel loop
                break;
            }
        } // while

        // Stop image acquisition
        Camera.AcquisitionStop.Execute();

        //
        // Cleanup
        //

        // Cleanup of image buffers
        // It is necessary to deregister a buffer before freeing the memory
        for (std::vector<CGrabBuffer*>::iterator it = BufferList.begin(); it != BufferList.end(); it++)
        {
            // deregister buffer
            StreamGrabber.DeregisterBuffer((*it)->GetBufferHandle());
            // delete buffer
            delete *it;
        }

        // cleanup of stream grabber
        StreamGrabber.FinishGrab();
        StreamGrabber.Close();

        // disable sending of end-of-exposure events
        Camera.EventSelector = EventSelector_ExposureEnd;
        Camera.EventNotification.SetValue(EventNotification_Off);

        // Cleanup of event grabber and event adapter
        // Deregister the callback first, otherwise when shutting down the event grabber
        Camera.ExposureEndEventTimestamp.GetNode()->DeregisterCallback(hCb);
        EventGrabber.Close();
        Camera.DestroyEventAdapter(pEventAdapter);


        // cleanup of camera object
        Camera.Close();
    }
    catch (GenICam::GenericException &e)
    {
        // Error handling
        cerr << "An exception occurred!" << endl
        << e.GetDescription() << endl;
        return 1;
    }
    // Quit application
    return 0;
}

