// This sample program demonstrates the usage of the Lookup Table feature
//



// Include files to use the PYLON API
#include <pylon/PylonIncludes.h>
using namespace Pylon;

#if defined( USE_1394 )
// settings to use  Basler 1394 cameras
#include <pylon/1394/Basler1394Camera.h>
typedef Pylon::CBasler1394Camera Camera_t;
using namespace Basler_IIDC1394CameraParams;
using namespace Basler_IIDC1394StreamParams;
#elif defined ( USE_GIGE )
// settings to use Basler GigE cameras
#include <pylon/gige/BaslerGigECamera.h>
typedef Pylon::CBaslerGigECamera Camera_t;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;
#else
#error Camera type is not specified. For example, define USE_GIGE for using GigE cameras
#endif

#pragma intrinsic(_byteswap_ulong)

// namespace for using cout
using namespace std;

int main(int argc, char* argv[])
{
    // Automagically call PylonInitialize and PylonTerminate to ensure the pylon runtime system
    // is initialized during lifetime of this object.
    Pylon::PylonAutoInitTerm autoInitTerm;

    try
    {
        // Get the transport layer factory
        CTlFactory& TlFactory = CTlFactory::GetInstance();

        // Create the transport layer object needed to enumerate or
        // create a camera object of type Camera_t::DeviceClass()
        ITransportLayer *pTl = TlFactory.CreateTl( Camera_t::DeviceClass() );

        // Exit application if the specific transport layer is not available
        if ( ! pTl )
        {
            cerr << "Failed to create transport layer!" << endl;
            return 1;
        }

        // Get all attached cameras and exit application if no camera is found
        DeviceInfoList_t devices;
        if ( 0 == pTl->EnumerateDevices( devices ) )
        {
            cerr << "No camera present!" << endl;
            return 1;
        }

        // Create the camera object of the first available camera
        Camera_t Camera(pTl->CreateDevice( devices[ 0 ] ));

        cout << "Opening camera...";

        // Open the camera
        Camera.Open();

        cout << "done" << endl;

        cout << "Writing LUT....";

        // Select the lookup table using the LUTSelector
        Camera.LUTSelector.SetValue( LUTSelector_Luminance );

        // Some cameras have 10 bit, other 12 bit lookup tables, so determine
        // hte type of the lookup table of the current device
        const int nValues = (int) Camera.LUTIndex.GetMax() + 1;
        int inc;
        if ( nValues == 4096 ) // 12 bit LUT
            inc = 8;
        else if ( nValues == 1024 ) // 10 bit LUT
            inc = 2;
        else
        {
            cerr << "Type of LUT is not supported by this sample!" << endl;
            return 1;
        }

        // Use LUTIndex and LUTValue parameter to access the lookup table values.
        // The following lookup table causes an inversion of the sensor values.

        for ( int i = 0; i < nValues; i += inc )
        {
            Camera.LUTIndex.SetValue( i );
            Camera.LUTValue.SetValue( nValues - 1 - i );
        }

        cout << "done" << endl;

        // Enable the lookup table
        Camera.LUTEnable.SetValue( true );

        // Grab and process images here
        // ...

        // Disable the lookup table
        Camera.LUTEnable.SetValue( false );

        // Close Camera
        Camera.Close();

    }
    catch( GenICam::GenericException &e )
    {
        // Error handling
        cerr << "An exception occurred!" << endl << e.GetDescription() << endl;
        return 1;
    }

    // Quit application
    return 0;
}

