// CGrabThread.cpp
//

#include "CGrabThread.h"

// Use PYLON related namespaces
using namespace Pylon;
using namespace GenICam;
using namespace GenApi;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;
using namespace Basler_GigETLParams;

using namespace std;

class CGrabBuffer
{
public:
    CGrabBuffer(const size_t ImageSize, Camera_t::StreamGrabber_t * pStreamGrabber);
    ~CGrabBuffer();
    uint8_t* GetBufferPointer( void ) { return m_pBuffer; }
    StreamBufferHandle GetBufferHandle( void ) { return m_hBuffer; }
    void RegisterBuffer( void );
    void DeregisterBuffer( void );

protected:
    uint8_t *m_pBuffer;
    Camera_t::StreamGrabber_t * m_pStreamGrabber;
    StreamBufferHandle m_hBuffer;
    const size_t m_ImageSize;
};

// Constructor allocates the image buffer
CGrabBuffer::CGrabBuffer(const size_t ImageSize, Camera_t::StreamGrabber_t * pStreamGrabber):
m_pBuffer( NULL ),
m_hBuffer( NULL ),
m_ImageSize( ImageSize ),
m_pStreamGrabber ( pStreamGrabber )
{
    m_pBuffer = new uint8_t[ ImageSize ];
    if ( NULL == m_pBuffer )
    {
        GenericException e("Not enough memory to allocate image buffer", __FILE__,__LINE__);
        throw e;
    }
}

// Buffers used for grabbing must be registered at the stream grabber.
// The registration returns a handle to be used for queuing the buffer
void CGrabBuffer::RegisterBuffer( void )
{
    m_hBuffer = m_pStreamGrabber->RegisterBuffer( m_pBuffer, m_ImageSize );
}

// It is necessary to deregister buffer before freeing the memory
void CGrabBuffer::DeregisterBuffer( void )
{
    if ( m_hBuffer )
        m_pStreamGrabber->DeregisterBuffer( m_hBuffer );
}

// Freeing the memory
CGrabBuffer::~CGrabBuffer()
{
    if ( NULL != m_pBuffer )
        delete[] m_pBuffer;
}

CGrabThread::CGrabThread( ThreadParams_t Threadparams)
:
m_Tp(Threadparams)
{
}

CGrabThread::~CGrabThread()
{
}


void CGrabThread::Create()
{
    pthread_create(&m_Thread, NULL, ThreadMain, (void*) this);

    return;
}


/// The thread function
void * CGrabThread::Run()
{
    try
    {
        // Open the camera
        if ( false == m_Tp.pCamera->IsOpen() )
            m_Tp.pCamera->Open();

        // Get the first stream grabber object of the selected camera
        Camera_t::StreamGrabber_t StreamGrabber(m_Tp.pCamera->GetStreamGrabber( 0 ));

        // Open the stream grabber
        StreamGrabber.Open();

        // Create an image buffer
        const size_t ImageSize = (size_t) ( m_Tp.pCamera->PayloadSize.GetValue() );

        // We won't use image buffers greater than ImageSize
        StreamGrabber.MaxBufferSize.SetValue(ImageSize);

        // We won't queue more than ten image buffers at a time
        StreamGrabber.MaxNumBuffer.SetValue( m_Tp.maxQueueBuffer);
        
        // Allocate all resources for grabbing. Critical parameters like image
        // size now could not be changed until FinishGrab() is called.
        StreamGrabber.PrepareGrab();

        // Buffers used for grabbing must be registered at the stream grabber.
        // The registration returns a handle to be used for queuing the buffer
        std::vector<CGrabBuffer*> BufferList;
        for (uint32_t i=0; i<m_Tp.maxQueueBuffer; ++i)
        {
            CGrabBuffer *pGrabBuffer = new CGrabBuffer( ImageSize, &StreamGrabber );
            pGrabBuffer->RegisterBuffer();

            // Put buffer into the grab queue for grabbing
            StreamGrabber.QueueBuffer( pGrabBuffer->GetBufferHandle(), NULL );

            // Put the grab buffer object into the buffer list
            BufferList.push_back( pGrabBuffer );
        }

        // This grab thread is ready for acquisition
        m_Tp.pReadyToGo->Signal( );

        // Wait 30 sec for starting all cameras simultaneously
        m_Tp.pGo->Wait( 30000 );
        {
            AutoLock lock(*m_Tp.pLock);
            cout << m_Tp.pDescription->str() << "Start grabbing ..." << endl;
        }

        {
            // Start grabbing
            uint32_t numImagesGrabbed = Grab( m_Tp.ImagesToGrab, m_Tp.maxQueueBuffer, &StreamGrabber );


            AutoLock lock(*m_Tp.pLock);

            cout << m_Tp.pDescription->str() << dec <<  "Images grabbed: "
                << numImagesGrabbed << endl;
        }

        // clean up

        // It is necessary to deregister buffer before freeing the memory
        for ( std::vector<CGrabBuffer*>::iterator it = BufferList.begin(); it != BufferList.end(); it++)
        {
            (*it)->DeregisterBuffer();
            delete *it;
        }

        // Free all resources used for grabbing
        StreamGrabber.FinishGrab();

        // Close stream grabber
        StreamGrabber.Close();
        {
            AutoLock lock(*m_Tp.pLock);
            cout << m_Tp.pDescription->str() << "Grabbing finished!" << endl;
        }
    }
    catch ( std::exception e )
    {
        AutoLock lock(*m_Tp.pLock);
        cout << m_Tp.pDescription->str() <<" CGrabThread"
            << " has caught exception and terminates." << e.what() << endl;
    }
    return 0;
}

uint32_t CGrabThread::Grab( const uint32_t numberOfImagesToGrab, const uint32_t maxQueueBuffer, Camera_t::StreamGrabber_t* pStreamGrabber)
{
    uint32_t grabbedImages = 0;
    uint32_t failedImages = 0;

    // Let the camera acquire images
    m_Tp.pCamera->AcquisitionStart.Execute();

    // Grab numberOfImagesToGrab times
    for ( uint32_t n = 0; n < numberOfImagesToGrab; n++)
    {
        // Wait for the grabbed image with timeout of 3 seconds
        if ( pStreamGrabber->GetWaitObject().Wait( 3000 ))
        {
            // Get the grab result from the grabber's result queue
            GrabResult Result;
            pStreamGrabber->RetrieveResult( Result );

            if ( Grabbed == Result.Status() )
            {
                grabbedImages++;

                // uncomment the following lines, if you want to get status output during image grab 
                // {
                //     AutoLock lock(*m_Tp.pLock);
                //     cout << m_Tp.pDescription->str() <<" GrabbedImages: " << grabbedImages << endl;
                // }

                // Image processing
                uint8_t * pImageBuffer = (uint8_t*) Result.Buffer();
                {
                    // Do what you need to do with the grabbed image!
                }
                // Reuse the buffer for grabbing the next image
                if ( n < numberOfImagesToGrab - maxQueueBuffer )
                    pStreamGrabber->QueueBuffer( Result.Handle(), NULL );
            }
            else if ( Failed == Result.Status() )
            {
                // Error Handling
                AutoLock lock(*m_Tp.pLock);
                failedImages++;

                cerr << m_Tp.pDescription->str() << "No image acquired!" << endl;
                cerr << m_Tp.pDescription->str() << "Error code : 0x" << hex
                     << Result.GetErrorCode() << endl;
                cerr << m_Tp.pDescription->str() <<"Error description : "
                     << Result.GetErrorDescription() << endl;

                // Reuse the buffer for grabbing the next image
                if ( n < numberOfImagesToGrab - maxQueueBuffer )
                    pStreamGrabber->QueueBuffer( Result.Handle(), NULL );
            }
        }
        else
        {
            // Timeout
            AutoLock lock(*m_Tp.pLock);
            cerr << m_Tp.pDescription->str() << "Timeout occurred!" << endl;

            // Get the pending buffer back (It is not allowed to deregister
            // buffers when they are still queued)
            pStreamGrabber->CancelGrab();

            // Get all buffers back
            for ( GrabResult r; pStreamGrabber->RetrieveResult( r ););

            // Cancel loop;
            break;
        }
    }
    {
        AutoLock lock(*m_Tp.pLock);
        cout << m_Tp.pDescription->str() << "Number of succesfuly grabbed images : "
            << grabbedImages << endl;
        cout << m_Tp.pDescription->str() << "Number of failed images             : "
            << failedImages << endl;
    }

    // Stop acquisition
    m_Tp.pCamera->AcquisitionStop.Execute();

    return grabbedImages + failedImages;
}
