// ManyCamerasUnix.cpp
//

// Include to use thread based Grabber, this includes also the pylon API
#include "CGrabThread.h"

// Settings to use Basler GigE cameras
#include <pylon/gige/BaslerGigECamera.h>

// Use PYLON related namespaces
using namespace Pylon;
using namespace GenICam;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;
using namespace Basler_GigETLParams;

using namespace std;

// Byte overhead of each packet
#define PROTOCOLOVERHEAD 26
// Buffers for grabbing
const uint32_t c_nBuffers = 10;
// Num cameras grab simultanously
const uint32_t c_MaxCameras = 3;
// Number of images to be grabbed
const uint32_t c_ImagesToGrab = 1000; 
// Packet size in bytes. In order to use values greater than 1500,
// jumbo frames at the network adapter must be enabled.
const uint32_t c_PacketSize = 1500;


int main()
{
    try
    {
        //  List of cameras
        DeviceInfoList_t m_Devices; // Still empty

        //  Create the unique transport layer factory object
        //  ... create the transport layer object
        //  ... create and initialise Cameras
        CTlFactory& TlFactory = CTlFactory::GetInstance();
        ITransportLayer *pTl = TlFactory.CreateTl( Camera_t::DeviceClass() );
        if ( ! pTl )
        {
            throw GenericException( 
                "Failed to create transport layer!", __FILE__, __LINE__);
        }

        // Get all attached cameras and exit if no camera is found
        m_Devices.clear();
        if ( 0 == pTl->EnumerateDevices( m_Devices ) )
        {
            throw GenericException( 
                "No camera present!", __FILE__, __LINE__);
        }
        cout << "Number of cameras found: "
             << m_Devices.size() << endl;

        // Array for variable number of cameras and camera grab thread objects
        Camera_t *m_pCameras[c_MaxCameras];
        CGrabThread *m_pGrabThreads[c_MaxCameras];

        // Create thread parameter object container
        ThreadParams_t Tp[c_MaxCameras];

        // Container for wait objects for camera signals
        WaitObjects AllCameras_ReadyToGo;

        // Create camera synchronisation objects list
        // Camera threads will signal when camera is ready for grab
        WaitObjectEx readyToGo[c_MaxCameras]; 
        
        // We signal camera threads to start grabbing
        WaitObjectEx go[c_MaxCameras];

        for (int i=0; i<c_MaxCameras; i++)
        {
            readyToGo[i] = WaitObjectEx::Create();
            go[i]        = WaitObjectEx::Create();
        }

        // Restrict number of use camera
        size_t nCameras = m_Devices.size();
        if ( nCameras >= c_MaxCameras )
            nCameras = c_MaxCameras;


        // Create a lock object to synchronize cout within the threads
        Pylon::CLock *pLock = new Pylon::CLock();

        // Create all camera devices and set up the grab threads
        for( uint32_t iCam=0; iCam<nCameras; ++iCam )
        {
            Camera_t *pCamera = new Camera_t( pTl->CreateDevice( m_Devices[iCam]) );
            if ( !pCamera )
            {
                throw GenericException (
                    "This should never happen: Camera list corrupted",
                    __FILE__, __LINE__ );
            }

            // Add camera to list
            m_pCameras[iCam] = pCamera; 

            // Initial clear
            m_pGrabThreads[iCam]=NULL;

            // Add camera synchronisation objects list
            AllCameras_ReadyToGo.Add( readyToGo[iCam] );

            // Set thread parameter for the current camera object
            Tp[iCam].maxQueueBuffer = c_nBuffers;
            Tp[iCam].ImagesToGrab = c_ImagesToGrab;
            Tp[iCam].pLock = pLock;
            Tp[iCam].pCamera = m_pCameras[iCam];
            Tp[iCam].pReadyToGo = &readyToGo[iCam];
            Tp[iCam].pGo = &go[iCam];

            // Set a description, this is used as prefix of each thread output
            Tp[iCam].pDescription = new ostringstream( ostringstream::out );
            *Tp[iCam].pDescription << "Thread #" << iCam << " : ";

            // Create the grab thread object
            m_pGrabThreads[iCam] = new CGrabThread( Tp[iCam] );

            if ( !m_pGrabThreads[iCam] ) 
                throw GenericException( 
                "Create CGrabThread failed!", __FILE__, __LINE__);
        }

        // Parametrize all cameras and start grab thread
        for( uint32_t iCam=0; iCam<nCameras; ++iCam )
        {
            Camera_t *pCamera = m_pCameras[iCam];

            // Open the camera
            pCamera->Open();

            // Set image properties
            pCamera->PixelFormat.SetValue( PixelFormat_Mono8 );
            pCamera->OffsetX.SetValue( 0 );
            pCamera->OffsetY.SetValue( 0 );
            pCamera->Width.SetValue( pCamera->Width.GetMax() );
            pCamera->Height.SetValue( pCamera->Height.GetMax() );

            // Camera settings to acquire continously images in free run mode
            pCamera->TriggerSelector.SetValue( TriggerSelector_AcquisitionStart );
            pCamera->TriggerMode.SetValue( TriggerMode_Off );
            pCamera->AcquisitionMode.SetValue( AcquisitionMode_Continuous );
            pCamera->ExposureMode.SetValue( ExposureMode_Timed );
            pCamera->ExposureTimeRaw.SetValue( pCamera->ExposureTimeRaw.GetMin() );

            // Transport layer parameter
            // Set Packetsize
            pCamera->GevSCPSPacketSize.SetValue( c_PacketSize );

            // Reduce max. required bandwith by setting the "inter-packet delay" and the "frame transfer start delay" transport layer parameter
            pCamera->GevSCPD.SetValue( (c_PacketSize + PROTOCOLOVERHEAD)* (nCameras-1) );
            pCamera->GevSCFTD.SetValue( (c_PacketSize + PROTOCOLOVERHEAD) * iCam );

            // Initial create
            m_pGrabThreads[iCam]->Create();
        }

        // Wait until all cameras are ready for grabbing
        AllCameras_ReadyToGo.WaitForAll( 10000 );

        // Start grabbing on all cameras
        for( uint32_t iCam=0; iCam<nCameras; ++iCam )
            go[iCam].Signal();

        // Wait until all threads have been terminated
        for( uint32_t iCam=0; iCam<nCameras; ++iCam )
            pthread_join(*m_pGrabThreads[iCam], NULL);

        //  Destroy all camera devices and threads
        for( uint32_t iCam=0; iCam<nCameras; ++iCam )
        {
            m_pCameras[iCam]->Close();
            delete m_pGrabThreads[iCam];
            delete Tp[iCam].pDescription;
            delete m_pCameras[iCam];
        }
    }
    catch( GenericException &e )
    {
        cerr << "****  An exception occurred! Desription is: " << endl 
            << "    " << e.GetDescription() << endl;
        return 1;
    }

    // Quit application
    return 0;
}
