// ParametrizeCamera.cpp
//


// include files to use the PYLON API
#include <pylon/PylonIncludes.h>
using namespace Pylon;

#if defined( USE_1394 )
// settings to use  Basler 1394 cameras
#include <pylon/1394/Basler1394Camera.h>
typedef Pylon::CBasler1394Camera Camera_t;
using namespace Basler_IIDC1394CameraParams;
using namespace Basler_IIDC1394StreamParams;
#elif defined ( USE_GIGE )
// settings to use Basler GigE cameras
#include <pylon/gige/BaslerGigECamera.h>
typedef Pylon::CBaslerGigECamera Camera_t;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;
#else
#error Camera type is not specified. For example, define USE_GIGE for using GigE cameras
#endif


// namespace for using cout
#include <iomanip>
using namespace std;

int64_t Adjust(int64_t v, const int64_t i)
{
    if (i == 1)
    {
        return v;
    }
    else if (i > 1)
    {
        int64_t r = (v + (i - 1)) / i;
        return r*i;
    }
    else
    {
        throw LOGICAL_ERROR_EXCEPTION("Unexpected increment %d", i);
    }

}
int main(int argc, char* argv[])
{
    // Automagically call PylonInitialize and PylonTerminate to ensure the pylon runtime system
    // is initialized during lifetime of this object.
    Pylon::PylonAutoInitTerm autoInitTerm;

    try
    {
        // Get the transport layer factory
        CTlFactory& TlFactory = CTlFactory::GetInstance();

        // Create the transport layer object needed to enumerate or
        // create a camera object of type Camera_t::DeviceClass()
        ITransportLayer *pTl = TlFactory.CreateTl(Camera_t::DeviceClass());

        // Exit application if the specific transport layer is not available
        if (! pTl)
        {
            cerr << "Failed to create transport layer!" << endl;
            return 1;
        }

        // Get all attached cameras and exit application if no camera is found
        DeviceInfoList_t devices;
        if (0 == pTl->EnumerateDevices(devices))
        {
            cerr << "No camera present!" << endl;
            return 1;
        }

        // Create the camera object of the first available camera
        // The camera object is used to set and get all available
        // camera features.
        Camera_t Camera(pTl->CreateDevice(devices[ 0 ]));

        // Get the first stream grabber object of the selected camera
        Camera_t::StreamGrabber_t StreamGrabber(Camera.GetStreamGrabber(0));

        // Open the camera
        Camera.Open();

        // Get camera device information
        cout << "Camera Device Information" << endl
        << "=========================" << endl;
        cout << "Vendor           : "
        << Camera.DeviceVendorName.GetValue() << endl;
        cout << "Model            : "
        << Camera.DeviceModelName.GetValue() << endl;
        cout << "Firmware version : "
        << Camera.DeviceFirmwareVersion.GetValue() << endl << endl;

        // Camera settings
        cout << "Camera Device Settings" << endl
        << "======================" << endl;
        // Set image format
        Camera.PixelFormat.SetValue(PixelFormat_Mono8);

        // Set AOI to half resolution
        Camera.Width.SetValue(Adjust(Camera.SensorWidth.GetValue() / 2, Camera.Width.GetInc()));
        Camera.Height.SetValue(Adjust(Camera.SensorHeight.GetValue() / 2, Camera.Height.GetInc()));
        cout << "Width            : " << Camera.Width.GetValue() << endl;
        cout << "Height           : " << Camera.Height.GetValue() << endl;

        // Set AOI to center
        Camera.OffsetX.SetValue(Adjust(Camera.OffsetX.GetMax() / 2, Camera.OffsetX.GetInc()));
        Camera.OffsetY.SetValue(Adjust(Camera.OffsetY.GetMax() / 2, Camera.OffsetY.GetInc()));
        cout << "OffsetX          : " << Camera.OffsetX.GetValue() << endl;
        cout << "OffsetY          : " << Camera.OffsetY.GetValue() << endl;

        // Set minimum gain
        Camera.GainRaw.SetValue(Camera.GainRaw.GetMin());
        cout << "Gain min         : " << Camera.GainRaw.GetValue() << endl;

        // Set maximum gain
        Camera.BlackLevelRaw.SetValue(Camera.BlackLevelRaw.GetMax());
        cout << "Black level max  : "
        << Camera.BlackLevelRaw.GetValue() << endl;

#if 1
        // Set white balance for red channel
        if (GenApi::IsAvailable(Camera.BalanceRatioAbs) &&
                GenApi::IsAvailable(Camera.BalanceRatioSelector))
        {
            Camera.BalanceRatioSelector.SetValue(BalanceRatioSelector_Red);
            Camera.BalanceRatioAbs.SetValue(Camera.BalanceRatioAbs.GetMax());
            cout << "Balance for "
            << setiosflags(ios::left) << setw(5)
            << Camera.BalanceRatioSelector.ToString()
            << resetiosflags(ios::left)
            << ": "
            << setprecision(3)
            << Camera.BalanceRatioAbs.GetValue()
            << endl;
        }
#endif

        // Close Camera
        Camera.Close();

    }
    catch (GenICam::GenericException &e)
    {
        // Error handling
        cerr << "An exception occurred!" << endl << e.GetDescription() << endl;
        return 1;
    }

    // Quit application
    return 0;
}

