// UserSets.cpp
//

// include files to use the PYLON API
#include <pylon/PylonIncludes.h>
using namespace Pylon;


#if defined( USE_1394 )
// settings to use  Basler 1394 cameras
#include <pylon/1394/Basler1394Camera.h>
typedef Pylon::CBasler1394Camera Camera_t;
using namespace Basler_IIDC1394CameraParams;
using namespace Basler_IIDC1394StreamParams;
#elif defined ( USE_GIGE )
// settings to use Basler GigE cameras
#include <pylon/gige/BaslerGigECamera.h>
typedef Pylon::CBaslerGigECamera Camera_t;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;
#else
#error Camera type is not specified. For example, define USE_GIGE for using GigE cameras
#endif



// namespace for using cout
using namespace std;


int main(int argc, char* argv[])
{
    // Automagically call PylonInitialize and PylonTerminate to ensure the pylon runtime system
    // is initialized during lifetime of this object.
    Pylon::PylonAutoInitTerm autoInitTerm;


    try
    {
        // Get the transport layer factory
        CTlFactory& TlFactory = CTlFactory::GetInstance();

        // Create the transport layer object needed to enumerate or
        // create a camera object of type Camera_t::DeviceClass()
        ITransportLayer *pTl = TlFactory.CreateTl(Camera_t::DeviceClass());

        // Exit application if the specific transport layer is not available
        if (! pTl)
        {
            cerr << "Failed to create transport layer!" << endl;
            return 1;
        }

        // Get all attached cameras and exit application if no camera is found
        DeviceInfoList_t devices;
        if (0 == pTl->EnumerateDevices(devices))
        {
            cerr << "No camera present!" << endl;
            return 1;
        }

        // Create the camera object of the first available camera
        // The camera object is used to set and get all available
        // camera features.
        Camera_t Camera(pTl->CreateDevice(devices[0]));

        // Get the first stream grabber object of the selected camera
        Camera_t::StreamGrabber_t StreamGrabber(Camera.GetStreamGrabber(0));

        // Open the camera
        Camera.Open();

        // Load default settings
        Camera.UserSetSelector.SetValue(UserSetSelector_Default);
        Camera.UserSetLoad.Execute();

        // Set minimum values
        Camera.GainRaw.SetValue(Camera.GainRaw.GetMin());
        Camera.BlackLevelRaw.SetValue(Camera.BlackLevelRaw.GetMin());
        Camera.ExposureTimeRaw.SetValue(Camera.ExposureTimeRaw.GetMin());

        // Store to user set 1
        Camera.UserSetSelector.SetValue(UserSetSelector_UserSet1);
        Camera.UserSetSave.Execute();

        // Show default setting
        Camera.UserSetSelector.SetValue(UserSetSelector_Default);
        Camera.UserSetLoad.Execute();
        cout << "Default settings" << endl;
        cout << "================" << endl;
        cout << "Gain          : 0x" << hex << Camera.GainRaw.GetValue() << endl;
        cout << "Black level   : 0x" << hex << Camera.BlackLevelRaw.GetValue() << endl;
        cout << "Exposure time : 0x" << hex << Camera.ExposureTimeRaw.GetValue() << endl;

        // Show default setting
        Camera.UserSetSelector.SetValue(UserSetSelector_UserSet1);
        Camera.UserSetLoad.Execute();
        cout << "UserSet1 settings" << endl;
        cout << "=================" << endl;
        cout << "Gain          : 0x" << hex << Camera.GainRaw.GetValue() << endl;
        cout << "Black level   : 0x" << hex << Camera.BlackLevelRaw.GetValue() << endl;
        cout << "Exposure time : 0x" << hex << Camera.ExposureTimeRaw.GetValue() << endl;

        // Set userset 1 as default userset
        Camera.UserSetDefaultSelector.SetValue(UserSetDefaultSelector_UserSet1);

        // Close Camera
        Camera.Close();

    }
    catch (GenICam::GenericException &e)
    {
        // Error handling
        cerr << "An exception occurred!" << endl << e.GetDescription() << endl;
        return 1;
    }

    // Quit application
    return 0;
}

