// WaitEx.cpp	
//

// This sample demonstrates how a grab loop can be terminated by a signal.

// For demonstration purposes, this sample generates the signal by using a realtime timer.
// When the timer expires, it creates a user signal that is handled within the grab loop. 
//
// The WaitEx() method is used to wait for buffers to be grabbed. The method returns either when 
// the timeout expires, a buffer is grabbed, or a signal occurred.

#include <pylon/PylonIncludes.h>  // include file to use the PYLON API

#include <iostream>
#include <signal.h>
#include <time.h>

#ifndef PYLON_LINUX_BUILD
#error This sample is currently only supported for Pylon4Linux
#endif

using namespace Pylon;

#include <pylon/gige/BaslerGigECamera.h>
typedef Pylon::CBaslerGigECamera Camera_t;
using namespace Basler_GigECameraParams;
using namespace Basler_GigEStreamParams;

using namespace std; // namespace for using cout


void dummyHandler(int signum)
{
    // do nothing
}

// This function installs a signal handler that does nothing.
// Installing a signal handler is necessary, otherwise the default handler would terminate the application.
void installSigHandler()
{
    struct sigaction act = {0};
    act.sa_handler = dummyHandler;
    if ( sigaction( SIGRTMIN, &act, NULL ) !=  0 ) 
        throw RUNTIME_EXCEPTION("Failed to install signal handler");
}


// Creates a timer that will issue a signal.
timer_t startTimer()
{
    timer_t id;
    struct sigevent sig = {0};
    sig.sigev_notify = SIGEV_SIGNAL;  // the timer issues a signal
    sig.sigev_signo = SIGRTMIN;       // the SIGRTMIN signal is issued

    if ( 0 != timer_create( CLOCK_REALTIME, &sig, &id ) )  // create the timer
        throw RUNTIME_EXCEPTION("Failed to create timer");


    // set the timer value
    struct itimerspec tspec = {0};
    tspec.it_interval.tv_sec = 0;   // one shot timer
    tspec.it_interval.tv_nsec = 0;  
    tspec.it_value.tv_sec = 3;  // timer expires after 3 seconds
    tspec.it_value.tv_nsec = 0;
    if ( 0 != timer_settime( id, 0, &tspec, NULL ) )
    {
        timer_delete( id );
        throw RUNTIME_EXCEPTION("Failed to parameterize timer");
    }

    return id;
}


// Simple class used for managing a memory buffer used for grabbing
class CGrabBuffer
{
    public:
        CGrabBuffer(const size_t ImageSize);
        ~CGrabBuffer();
        uint8_t* GetBufferPointer(void) { return m_pBuffer; }
        StreamBufferHandle GetBufferHandle(void) { return m_hBuffer; }
        void SetBufferHandle(StreamBufferHandle hBuffer) { m_hBuffer = hBuffer; };

    protected:
        uint8_t *m_pBuffer;
        StreamBufferHandle m_hBuffer;
};

// Constructor allocates the image buffer
CGrabBuffer::CGrabBuffer(const size_t ImageSize):
        m_pBuffer(NULL)
{
    m_pBuffer = new uint8_t[ ImageSize ];
    if (NULL == m_pBuffer)
    {
        GenICam::GenericException e("Not enough memory to allocate image buffer", __FILE__, __LINE__);
        throw e;
    }
}

// Frees the memory
CGrabBuffer::~CGrabBuffer()
{
    if (NULL != m_pBuffer)
        delete[] m_pBuffer;
}


// Number of buffers used for grabbing
static const uint32_t c_nBuffers = 5;


int main(int argc, char* argv[])
{
    // Automagically call PylonInitialize and PylonTerminate to ensure the pylon runtime system
    // is initialized during lifetime of this object.
    Pylon::PylonAutoInitTerm autoInitTerm;

    try
    {
        // Get the transport layer factory
        CTlFactory& TlFactory = CTlFactory::GetInstance();

        // Create the transport layer object needed to enumerate or
        // create a camera object of type Camera_t::DeviceClass()
        ITransportLayer *pTl = TlFactory.CreateTl(Camera_t::DeviceClass());

        // Exit application if the specific transport layer is not available
        if (! pTl)
        {
            cerr << "Failed to create transport layer!" << endl;
            return 1;
        }

        // Get all attached cameras and exit application if no camera is found
        DeviceInfoList_t devices;
        if (0 == pTl->EnumerateDevices(devices))
        {
            cerr << "No camera present!" << endl;
            return 1;
        }

        // Create the camera object of the first available camera
        // The camera object is used to set and get all available
        // camera features.
        Camera_t Camera(pTl->CreateDevice(devices[ 0 ]));

        // Open the camera
        Camera.Open();

        // Get the first stream grabber object of the selected camera
        Camera_t::StreamGrabber_t StreamGrabber(Camera.GetStreamGrabber(0));

        // Open the stream grabber
        StreamGrabber.Open();

        // Set the camera to continuous frame mode
        Camera.TriggerSelector.SetValue(TriggerSelector_AcquisitionStart);
        Camera.TriggerMode.SetValue(TriggerMode_Off);
        Camera.AcquisitionMode.SetValue(AcquisitionMode_Continuous);
        Camera.ExposureMode.SetValue(ExposureMode_Timed);
        Camera.ExposureTimeRaw.SetValue(100);

        // Get the image buffer size
        const size_t ImageSize = (size_t)(Camera.PayloadSize.GetValue());

        // We won't use image buffers greater than ImageSize
        StreamGrabber.MaxBufferSize.SetValue(ImageSize);

        // We won't queue more than c_nBuffers image buffer at a time
        StreamGrabber.MaxNumBuffer.SetValue(c_nBuffers);

        // Allocate all resources for grabbing. Critical parameters like image
        // size now must not be changed until FinishGrab() is called.
        StreamGrabber.PrepareGrab();

        // Buffers used for grabbing must be registered at the stream grabber.
        // The registration returns a handle to be used for queuing the buffer
        std::vector<CGrabBuffer*> BufferList;
        for (uint32_t i = 0; i < c_nBuffers; ++i)
        {
            CGrabBuffer *pGrabBuffer = new CGrabBuffer(ImageSize);
            pGrabBuffer->SetBufferHandle(StreamGrabber.RegisterBuffer(
                pGrabBuffer->GetBufferPointer(), ImageSize));

            // Put the grab buffer object into the buffer list
            BufferList.push_back(pGrabBuffer);
        }

        for (std::vector<CGrabBuffer*>::const_iterator x = BufferList.begin(); x != BufferList.end(); ++x)
        {
            // Put buffer into the grab queue for grabbing
            StreamGrabber.QueueBuffer((*x)->GetBufferHandle(), NULL);
        }

        // It is necessary to install signal handlers for all signals that should be handled within
        // the grab loop. For this application, we are using the SIGRTMIN signal.
        installSigHandler();

        // starts a timer that will issue a SIGRTMIN signal when it expires
        timer_t timer = startTimer();

        // Let the camera acquire continuously images ( Acquisiton mode equals
        // Continuous! )
        Camera.AcquisitionStart.Execute();

        bool terminate = false;  // will be set to true when a signal has been detected
        // Grab images until we get a signal
        while ( ! terminate )
        {
            // Wait for the grabbed image with timeout of 10 seconds. We want to be interruptible by signals.
            waitex_result_t waitResult = StreamGrabber.GetWaitObject().WaitEx(10000, true);
            switch ( waitResult )
            {
                case waitex_timeout:  // timeout occurred, no buffer available
                    cerr << "Failed to grab image: timout" << endl;
                    continue;
                case waitex_signaled:  // buffer is available in the driver's output queue
                {
                    // Get the grab result 
                    GrabResult Result;
                    StreamGrabber.RetrieveResult(Result);

                    if (Grabbed == Result.Status())
                    {
                        // Grabbing was successful, process image
                        cout << "." << flush;
                    }
                    else if (Failed == Result.Status())
                    {
                        // Error Handling
                        cerr << "Failed to acquire image!" << endl;
                        cerr << "Error code : 0x" << hex
                        << Result.GetErrorCode() << endl;
                        cerr << "Error description : "
                        << Result.GetErrorDescription() << endl;
                    }

                    // Reuse the buffer for grabbing the next image
                    StreamGrabber.QueueBuffer(Result.Handle(), NULL);

                }
                break;
 
               case waitex_alerted:  // Wait operation has been interrupted by a signal
                    cout << endl << "Got signal. Terminating application" << endl;
                    terminate = true;
                    break;
            }
        }

        // Stop acquisition
        Camera.AcquisitionStop.Execute();


        // Cancel all pending buffers so that we can retrieve them from the output queue
        StreamGrabber.CancelGrab();

        // Get all buffers back. All buffers must be retrieved before they can be deregistered
       for (GrabResult r; StreamGrabber.RetrieveResult(r););

        // It is necessary to deregister buffer before freeing the memory
        for (std::vector<CGrabBuffer*>::iterator it = BufferList.begin(); it != BufferList.end(); it++)
        {
            StreamGrabber.DeregisterBuffer((*it)->GetBufferHandle());
            delete *it;
        }

        // Free all resources used for grabbing
        StreamGrabber.FinishGrab();

        // Close stream grabber
        StreamGrabber.Close();

        // Close camera
        Camera.Close();

        // Free the timer
        timer_delete( timer );

    }
    catch (GenICam::GenericException &e)
    {
        // Error handling
        cerr << "An exception occurred!" << endl
        << e.GetDescription() << endl;
        return 1;
    }
    // Quit application
    return 0;
}

